#include <math.h>
#include "SDL.h"
#include "bump.h"

static double light_map_p[LIGHT_SIZE][LIGHT_SIZE];
static double light_map_h[LIGHT_SIZE][LIGHT_SIZE];
static double light_map_v[LIGHT_SIZE][LIGHT_SIZE];

static inline Uint8 get_pixel8(SDL_Surface *s, int x, int y) {
  return ((Uint8 *)(s->pixels + y * s->pitch))[x];
}

static inline Uint32 get_pixel(SDL_Surface *s, int x, int y) {
  return ((Uint32 *)(s->pixels + y * s->pitch))[x];
}

static inline void set_pixel(SDL_Surface *s, int x, int y, Uint32 p) {
  ((Uint32 *)(s->pixels + y * s->pitch))[x] = p;
}

static inline int clip(int a) {
  if (a < 0) return 0;
  if (a > 255) return 255;
  return a;
}

void bump(SDL_Surface *map, SDL_Surface *texture, SDL_Surface *work,
	  int x, int y) {
  int i, j;
  for (i = 0; i < LIGHT_SIZE; i++)
    for (j = 0; j < LIGHT_SIZE; j++) {
      Uint32 vp, hp, p;
      int r, g, b;
      double br;
      vp = get_pixel8(map, x + j, y + i + 1);
      hp = get_pixel8(map, x + j + 1, y + i);
      p = get_pixel8(map, x + j, y + i);
      br = light_map_p[i][j] * 5
	+ light_map_v[i][j] * ((int)vp - (int)p)
	+ light_map_h[i][j] * ((int)hp - (int)p);

      p = get_pixel(texture, x + j, y + i);
      r = clip((p >> 16) * br);
      g = clip(((p >> 8) & 0xff) * br);
      b = clip((p & 0xff) * br);
      set_pixel(work, j, i, (r << 16) | (g << 8) | b);
    }
}

static inline double fmax(double a, double b) {
  return a > b ? a : b;
}

void init_bump(void) {
  int i, j;
  for (i = 0; i < LIGHT_SIZE; i++)
    for (j = 0; j < LIGHT_SIZE; j++) {
      double x, y, r, a;
      x = (double)j / (LIGHT_SIZE / 2) - 1;
      y = (double)i / (LIGHT_SIZE / 2) - 1;
      r = sqrt(x * x + y * y);
      light_map_h[i][j] = x * fmax(1 - r * r, 0);
      light_map_v[i][j] = y * fmax(1 - r * r, 0);
      light_map_p[i][j] = (1 - r) * fmax(1 - r * r, 0);
    }
}

